@@ -5,45 +5,70 @@
 
 use std::collections::HashMap;
 use std::fs::File;
-use std::fs::OpenOptions;
-use std::io::Write;
 use std::io::{self};
-use std::path::PathBuf;
-use std::process;
 
-extern crate colored;
 extern crate niffler;
 
-use anyhow::{anyhow, Result};
+use anyhow::{anyhow, Context, Result};
 use bio::io::{fasta, fastq};
-use colored::*;
 
-// msg function
+use crate::error;
 
-/// Print given text to stdout
-///
-/// # Example
-/// ```rust
-/// msg("Hello World!");
-/// ```
-///
-pub fn msg(msg: &str) {
-    writeln!(io::stdout(), "{}", format!("[INFO] {}", msg))
-        .expect("Cannot write to stdout");
+// to_niffler_format function
+pub fn to_niffler_format(format: &str) -> Result<niffler::compression::Format> {
+    if format == "gz" {
+        Ok(niffler::compression::Format::Gzip)
+    } else if format == "bz2" {
+        Ok(niffler::compression::Format::Bzip)
+    } else if format == "xz" {
+        Ok(niffler::compression::Format::Lzma)
+    } else {
+        Ok(niffler::compression::Format::No)
+    }
 }
 
-// err function
+// to_compression_ext function
 
-/// Print given text to stderr in red
+/// Convert niffler compression format to a file extension
 ///
 /// # Example
 /// ```rust
-/// err("This is an error");
+/// let compression = niffler::compression::Format::Gzip;
+/// let ext = to_compression_ext(compression);
+/// assert_eq!(ext, ".gz");
 /// ```
 ///
-pub fn err(error: &str) {
-    writeln!(io::stderr(), "{}", format!("[ERROR] {}", error).red())
-        .expect("Cannot write to stderr");
+pub fn to_compression_ext(compression: niffler::compression::Format) -> String {
+    match compression {
+        niffler::compression::Format::Gzip => ".gz".to_string(),
+        niffler::compression::Format::Bzip => ".bz2".to_string(),
+        niffler::compression::Format::Lzma => ".xz".to_string(),
+        niffler::compression::Format::No => "".to_string(),
+    }
+}
+
+// to_niffler_level function
+
+/// Convert an integer to a niffler::Level
+///
+/// # Example
+/// ```rust
+/// assert_eq!(to_niffler_level(3), niffler::Level::Three);
+/// ```
+///
+pub fn to_niffler_level(int_level: i32) -> niffler::Level {
+    match int_level {
+        1 => niffler::Level::One,
+        2 => niffler::Level::Two,
+        3 => niffler::Level::Three,
+        4 => niffler::Level::Four,
+        5 => niffler::Level::Five,
+        6 => niffler::Level::Six,
+        7 => niffler::Level::Seven,
+        8 => niffler::Level::Eight,
+        9 => niffler::Level::Nine,
+        _ => niffler::Level::One,
+    }
 }
 
 // read_file function -------------------------------------------------------
@@ -61,20 +86,15 @@ pub fn err(error: &str) {
 pub fn read_file(
     filename: &str,
 ) -> Result<(Box<dyn io::Read>, niffler::compression::Format)> {
-    let raw_in = Box::new(io::BufReader::new(File::open(filename)?));
+    let raw_in = Box::new(io::BufReader::new(
+        File::open(filename).with_context(|| error::Error::CantReadFile {
+            filename: filename.to_string(),
+        })?,
+    ));
 
-    let (reader, compression) =
-        niffler::get_reader(raw_in).expect("Cannot read input fasta file");
-
-    match compression {
-        niffler::compression::Format::Gzip => Ok((reader, compression)),
-        niffler::compression::Format::No => Ok((reader, compression)),
-        _ => {
-            writeln!(io::stderr(), "[ERROR] Only gzipped files are supported")
-                .expect("Cannot write to stderr");
-            process::exit(1);
-        }
-    }
+    niffler::get_reader(raw_in).with_context(|| {
+        anyhow!("Could not detect compression of file '{}'", filename)
+    })
 }
 
 /// get_file_type function --------------------------------------------------
@@ -84,6 +104,7 @@ pub fn read_file(
 pub enum FileType {
     Fasta,
     Fastq,
+    None,
 }
 
 /// Get file type from filename
@@ -94,17 +115,27 @@ pub enum FileType {
 /// let file_type = get_file_type(filename);
 /// ```
 ///
-pub fn get_file_type(filename: &str) -> Option<FileType> {
-    if filename.contains(".fastq") || filename.contains(".fq") {
-        Some(FileType::Fastq)
+pub fn get_file_type(
+    filename: &str,
+) -> Result<(FileType, niffler::compression::Format)> {
+    let mut filext = FileType::None;
+
+    if filext == FileType::None && filename.contains(".fastq")
+        || filename.contains(".fq")
+    {
+        filext = FileType::Fastq;
     } else if filename.contains(".fasta")
         || filename.contains(".fa")
         || filename.contains(".fas")
     {
-        Some(FileType::Fasta)
+        filext = FileType::Fasta;
     } else {
-        None
+        filext = FileType::None;
     }
+
+    let (_reader, comp) = read_file(filename)?;
+
+    Ok((filext, comp))
 }
 
 // split_line_by_tab function -----------------------------------------------
@@ -148,12 +179,9 @@ pub fn write_to_fa<'a>(
     file: &'a std::fs::File,
     compression: niffler::compression::Format,
     record: &'a fasta::Record,
+    level: niffler::Level,
 ) -> Result<()> {
-    let handle = niffler::get_writer(
-        Box::new(file),
-        compression,
-        niffler::compression::Level::One,
-    )?;
+    let handle = niffler::get_writer(Box::new(file), compression, level)?;
 
     let mut writer = fasta::Writer::new(handle);
     let _write_res = writer.write_record(&record)?;
@@ -177,15 +205,17 @@ pub fn write_to_fq<'a>(
     file: &'a std::fs::File,
     compression: niffler::compression::Format,
     record: &'a fastq::Record,
+    level: niffler::Level,
 ) -> Result<()> {
-    let handle = niffler::get_writer(
-        Box::new(file),
-        compression,
-        niffler::compression::Level::One,
-    )?;
+    let handle = niffler::get_writer(Box::new(file), compression, level)
+        .with_context(|| anyhow!("Could not get file writer"))?;
 
     let mut writer = fastq::Writer::new(handle);
-    let _write_res = writer.write_record(&record)?;
+    let _write_res = writer.write_record(&record).with_context(|| {
+        error::Error::CantWriteFile {
+            filename: "output file".to_string(),
+        }
+    })?;
 
     Ok(())
 }
@@ -235,32 +265,28 @@ pub fn bc_cmp(bc: &[u8], seq: &[u8], mismatch: i32) -> bool {
 ///
 ///
 pub fn se_fa_demux<'a>(
-    forward_records: &'a mut fasta::Records<std::boxed::Box<dyn std::io::Read>>,
-    compression: niffler::compression::Format,
+    forward: &'a str,
+    format: niffler::compression::Format,
+    level: niffler::Level,
     barcode_data: &'a Barcode,
     mismatch: i32,
-    out: &'a str,
     nb_records: &'a mut HashMap<&'a [u8], i32>,
-) -> Result<&'a mut HashMap<&'a [u8], i32>> {
-    let mut ext = "";
-    if compression == niffler::compression::Format::Gzip && ext.is_empty() {
-        ext = ".gz";
-    } else {
-        ext = "";
-    }
-
-    let mut unk_path = PathBuf::from("");
-    unk_path.push(out);
-    unk_path.push(format!("{}{}", "unknown.fa", ext));
-    let unknown_file = OpenOptions::new()
-        .create(true)
-        .append(true)
-        .open(unk_path)
-        .expect("Cannot create file");
+) -> Result<(&'a mut HashMap<&'a [u8], i32>, bool)> {
+    let (forward_reader, mut compression) =
+        read_file(forward).with_context(|| error::Error::ReadingError {
+            filename: forward.to_string(),
+        })?;
+    let mut forward_records = fasta::Reader::new(forward_reader).records();
 
     let my_vec = barcode_data.keys().cloned().collect::<Vec<_>>();
     let bc_len = my_vec[0].len();
 
+    let mut is_unk_empty = true;
+
+    if format != niffler::compression::Format::No {
+        compression = format;
+    }
+
     while let Some(Ok(f_rec)) = forward_records.next() {
         let mut iter = my_vec.iter();
         let res = iter.find(|&&x| bc_cmp(x, &f_rec.seq()[..bc_len], mismatch));
@@ -272,21 +298,32 @@ pub fn se_fa_demux<'a>(
                     &barcode_data.get(i).unwrap()[0],
                     compression,
                     &f_rec,
+                    level,
                 )
-                .expect("Cannot write to output file");
+                .with_context(|| {
+                    error::Error::WritingErrorNoFilename {
+                        format: FileType::Fasta,
+                    }
+                })?;
             }
             None => {
-                nb_records
-                    .entry(b"unknown.fq")
-                    .and_modify(|e| *e += 1)
-                    .or_insert(1);
-                write_to_fa(&unknown_file, compression, &f_rec)
-                    .expect("Cannot write to unknown file");
+                is_unk_empty = false;
+                write_to_fa(
+                    &barcode_data.get(&"XXX".as_bytes()).unwrap()[0],
+                    compression,
+                    &f_rec,
+                    level,
+                )
+                .with_context(|| {
+                    error::Error::WritingErrorNoFilename {
+                        format: FileType::Fasta,
+                    }
+                })?;
             }
         }
     }
 
-    Ok(nb_records)
+    Ok((nb_records, is_unk_empty))
 }
 
 // se_fq_demux function -----------------------------------------------------
@@ -309,32 +346,28 @@ pub fn se_fa_demux<'a>(
 ///
 ///
 pub fn se_fq_demux<'a>(
-    forward_records: &'a mut fastq::Records<std::boxed::Box<dyn std::io::Read>>,
-    compression: niffler::compression::Format,
+    forward: &'a str,
+    format: niffler::compression::Format,
+    level: niffler::Level,
     barcode_data: &'a Barcode,
     mismatch: i32,
-    out: &'a str,
     nb_records: &'a mut HashMap<&'a [u8], i32>,
-) -> Result<&'a mut HashMap<&'a [u8], i32>> {
-    let mut ext = "";
-    if compression == niffler::compression::Format::Gzip && ext.is_empty() {
-        ext = ".gz";
-    } else {
-        ext = "";
-    }
+) -> Result<(&'a mut HashMap<&'a [u8], i32>, bool)> {
+    let (forward_reader, mut compression) =
+        read_file(forward).with_context(|| error::Error::ReadingError {
+            filename: forward.to_string(),
+        })?;
+    let mut forward_records = fastq::Reader::new(forward_reader).records();
 
-    let mut unk_path = PathBuf::from("");
-    unk_path.push(out);
-    unk_path.push(format!("{}{}", "unknown.fq", ext));
-    let unknown_file = OpenOptions::new()
-        .create(true)
-        .append(true)
-        .open(unk_path)
-        .expect("Cannot create file");
+    if format != niffler::compression::Format::No {
+        compression = format;
+    }
 
     let my_vec = barcode_data.keys().cloned().collect::<Vec<_>>();
     let bc_len = my_vec[0].len();
 
+    let mut is_unk_empty = true;
+
     while let Some(Ok(f_rec)) = forward_records.next() {
         let mut iter = my_vec.iter();
         let res = iter.find(|&&x| bc_cmp(x, &f_rec.seq()[..bc_len], mismatch));
@@ -346,21 +379,32 @@ pub fn se_fq_demux<'a>(
                     &barcode_data.get(i).unwrap()[0],
                     compression,
                     &f_rec,
+                    level,
                 )
-                .expect("Cannot write to output file");
+                .with_context(|| {
+                    error::Error::WritingErrorNoFilename {
+                        format: FileType::Fastq,
+                    }
+                })?;
             }
             None => {
-                nb_records
-                    .entry(b"unknown.fq")
-                    .and_modify(|e| *e += 1)
-                    .or_insert(1);
-                write_to_fq(&unknown_file, compression, &f_rec)
-                    .expect("Cannot write to unknown file");
+                is_unk_empty = false;
+                write_to_fq(
+                    &barcode_data.get(&"XXX".as_bytes()).unwrap()[0],
+                    compression,
+                    &f_rec,
+                    level,
+                )
+                .with_context(|| {
+                    error::Error::WritingErrorNoFilename {
+                        format: FileType::Fastq,
+                    }
+                })?;
             }
         }
     }
 
-    Ok(nb_records)
+    Ok((nb_records, is_unk_empty))
 }
 
 // pe_fa_demux function -----------------------------------------------------
@@ -370,41 +414,36 @@ pub fn se_fq_demux<'a>(
 ///
 ///
 pub fn pe_fa_demux<'a>(
-    forward_records: &'a mut fasta::Records<std::boxed::Box<dyn std::io::Read>>,
-    reverse_records: &'a mut fasta::Records<std::boxed::Box<dyn std::io::Read>>,
-    compression: niffler::compression::Format,
+    forward: &'a str,
+    reverse: &'a str,
+    format: niffler::compression::Format,
+    level: niffler::Level,
     barcode_data: &'a Barcode,
     mismatch: i32,
-    out: &'a str,
     nb_records: &'a mut HashMap<&'a [u8], i32>,
-) -> Result<&'a mut HashMap<&'a [u8], i32>> {
-    let mut ext = "";
-    if compression == niffler::compression::Format::Gzip && ext.is_empty() {
-        ext = ".gz";
-    } else {
-        ext = "";
-    }
+) -> Result<(&'a mut HashMap<&'a [u8], i32>, String)> {
+    let (forward_reader, mut compression) =
+        read_file(forward).with_context(|| error::Error::ReadingError {
+            filename: forward.to_string(),
+        })?;
+    let mut forward_records = fasta::Reader::new(forward_reader).records();
 
-    let mut unk_path1 = PathBuf::from("");
-    unk_path1.push(out);
-    unk_path1.push(format!("{}{}", "unknown_R1.fa", ext));
-    let mut unk_path2 = PathBuf::from("");
-    unk_path2.push(out);
-    unk_path2.push(format!("{}{}", "unknown_R2.fa", ext));
-    let unknown_file1 = OpenOptions::new()
-        .create(true)
-        .append(true)
-        .open(unk_path1)
-        .expect("Cannot create file");
-    let unknown_file2 = OpenOptions::new()
-        .create(true)
-        .append(true)
-        .open(unk_path2)
-        .expect("Cannot create file");
+    let (reverse_reader, _compression) =
+        read_file(reverse).with_context(|| error::Error::ReadingError {
+            filename: reverse.to_string(),
+        })?;
+    let mut reverse_records = fasta::Reader::new(reverse_reader).records();
+
+    if format != niffler::compression::Format::No {
+        compression = format;
+    }
 
     let my_vec = barcode_data.keys().cloned().collect::<Vec<_>>();
     let bc_len = my_vec[0].len();
 
+    let mut unk1_empty = "true";
+    let mut unk2_empty = "true";
+
     while let Some(Ok(f_rec)) = forward_records.next() {
         let mut iter = my_vec.iter();
         let res = iter.find(|&&x| bc_cmp(x, &f_rec.seq()[..bc_len], mismatch));
@@ -416,16 +455,27 @@ pub fn pe_fa_demux<'a>(
                     &barcode_data.get(i).unwrap()[0],
                     compression,
                     &f_rec,
+                    level,
                 )
-                .expect("Cannot write to output file");
+                .with_context(|| {
+                    error::Error::WritingErrorNoFilename {
+                        format: FileType::Fasta,
+                    }
+                })?;
             }
             None => {
-                nb_records
-                    .entry(b"unknown_R1.fa")
-                    .and_modify(|e| *e += 1)
-                    .or_insert(1);
-                write_to_fa(&unknown_file1, compression, &f_rec)
-                    .expect("Cannot write to unknown file");
+                unk1_empty = "false";
+                write_to_fa(
+                    &barcode_data.get(&"XXX".as_bytes()).unwrap()[0],
+                    compression,
+                    &f_rec,
+                    level,
+                )
+                .with_context(|| {
+                    error::Error::WritingErrorNoFilename {
+                        format: FileType::Fasta,
+                    }
+                })?;
             }
         }
     }
@@ -441,21 +491,32 @@ pub fn pe_fa_demux<'a>(
                     &barcode_data.get(i).unwrap()[1],
                     compression,
                     &r_rec,
+                    level,
                 )
-                .expect("Cannot write to output file");
+                .with_context(|| {
+                    error::Error::WritingErrorNoFilename {
+                        format: FileType::Fasta,
+                    }
+                })?;
             }
             None => {
-                nb_records
-                    .entry(b"unknown_R2.fa")
-                    .and_modify(|e| *e += 1)
-                    .or_insert(1);
-                write_to_fa(&unknown_file2, compression, &r_rec)
-                    .expect("Cannot write to unknown file");
+                unk2_empty = "false";
+                write_to_fa(
+                    &barcode_data.get(&"XXX".as_bytes()).unwrap()[1],
+                    compression,
+                    &r_rec,
+                    level,
+                )
+                .with_context(|| {
+                    error::Error::WritingErrorNoFilename {
+                        format: FileType::Fasta,
+                    }
+                })?;
             }
         }
     }
-
-    Ok(nb_records)
+    let final_str = format!("{}{}", unk1_empty, unk2_empty);
+    Ok((nb_records, final_str))
 }
 
 // pe_fq_demux function -----------------------------------------------------
@@ -465,41 +526,36 @@ pub fn pe_fa_demux<'a>(
 ///
 ///
 pub fn pe_fq_demux<'a>(
-    forward_records: &'a mut fastq::Records<std::boxed::Box<dyn std::io::Read>>,
-    reverse_records: &'a mut fastq::Records<std::boxed::Box<dyn std::io::Read>>,
-    compression: niffler::compression::Format,
+    forward: &'a str,
+    reverse: &'a str,
+    format: niffler::compression::Format,
+    level: niffler::Level,
     barcode_data: &'a Barcode,
     mismatch: i32,
-    out: &'a str,
     nb_records: &'a mut HashMap<&'a [u8], i32>,
-) -> Result<&'a mut HashMap<&'a [u8], i32>> {
-    let mut ext = "";
-    if compression == niffler::compression::Format::Gzip && ext.is_empty() {
-        ext = ".gz";
-    } else {
-        ext = "";
-    }
+) -> Result<(&'a mut HashMap<&'a [u8], i32>, String)> {
+    let (forward_reader, mut compression) =
+        read_file(forward).with_context(|| error::Error::ReadingError {
+            filename: forward.to_string(),
+        })?;
+    let mut forward_records = fastq::Reader::new(forward_reader).records();
 
-    let mut unk_path1 = PathBuf::from("");
-    unk_path1.push(out);
-    unk_path1.push(format!("{}{}", "unknown_R1.fq", ext));
-    let mut unk_path2 = PathBuf::from("");
-    unk_path2.push(out);
-    unk_path2.push(format!("{}{}", "unknown_R2.fq", ext));
-    let unknown_file1 = OpenOptions::new()
-        .create(true)
-        .append(true)
-        .open(unk_path1)
-        .expect("Cannot create file");
-    let unknown_file2 = OpenOptions::new()
-        .create(true)
-        .append(true)
-        .open(unk_path2)
-        .expect("Cannot create file");
+    let (reverse_reader, _compression) =
+        read_file(reverse).with_context(|| error::Error::ReadingError {
+            filename: reverse.to_string(),
+        })?;
+    let mut reverse_records = fastq::Reader::new(reverse_reader).records();
+
+    if format != niffler::compression::Format::No {
+        compression = format;
+    }
 
     let my_vec = barcode_data.keys().cloned().collect::<Vec<_>>();
     let bc_len = my_vec[0].len();
 
+    let mut unk1_empty = "true";
+    let mut unk2_empty = "true";
+
     while let Some(Ok(f_rec)) = forward_records.next() {
         let mut iter = my_vec.iter();
         let res = iter.find(|&&x| bc_cmp(x, &f_rec.seq()[..bc_len], mismatch));
@@ -511,16 +567,27 @@ pub fn pe_fq_demux<'a>(
                     &barcode_data.get(i).unwrap()[0],
                     compression,
                     &f_rec,
+                    level,
                 )
-                .expect("Cannot write to output file");
+                .with_context(|| {
+                    error::Error::WritingErrorNoFilename {
+                        format: FileType::Fastq,
+                    }
+                })?;
             }
             None => {
-                nb_records
-                    .entry(b"unknown_R1.fq")
-                    .and_modify(|e| *e += 1)
-                    .or_insert(1);
-                write_to_fq(&unknown_file1, compression, &f_rec)
-                    .expect("Cannot write to unknown file");
+                unk1_empty = "false";
+                write_to_fq(
+                    &barcode_data.get(&"XXX".as_bytes()).unwrap()[0],
+                    compression,
+                    &f_rec,
+                    level,
+                )
+                .with_context(|| {
+                    error::Error::WritingErrorNoFilename {
+                        format: FileType::Fastq,
+                    }
+                })?;
             }
         }
     }
@@ -536,21 +603,33 @@ pub fn pe_fq_demux<'a>(
                     &barcode_data.get(i).unwrap()[1],
                     compression,
                     &r_rec,
+                    level,
                 )
-                .expect("Cannot write to output file");
+                .with_context(|| {
+                    error::Error::WritingErrorNoFilename {
+                        format: FileType::Fastq,
+                    }
+                })?;
             }
             None => {
-                nb_records
-                    .entry(b"unknown_R2.fq")
-                    .and_modify(|e| *e += 1)
-                    .or_insert(1);
-                write_to_fq(&unknown_file2, compression, &r_rec)
-                    .expect("Cannot write to unknown file");
+                unk2_empty = "false";
+                write_to_fq(
+                    &barcode_data.get(&"XXX".as_bytes()).unwrap()[1],
+                    compression,
+                    &r_rec,
+                    level,
+                )
+                .with_context(|| {
+                    error::Error::WritingErrorNoFilename {
+                        format: FileType::Fastq,
+                    }
+                })?;
             }
         }
     }
 
-    Ok(nb_records)
+    let final_str = format!("{}{}", unk1_empty, unk2_empty);
+    Ok((nb_records, final_str))
 }
 
 // Tests --------------------------------------------------------------------
@@ -621,31 +700,21 @@ mod tests {
     // se_fa_demux tests ----------------------------------------------------
     #[test]
     fn test_se_fa_demux() {
-        let out = "tests";
-        let (fr, cmp) =
-            read_file("tests/test2.fa.gz").expect("Cannot open file");
-        let mut records = fasta::Reader::new(fr).records();
         let mut bc_data: Barcode = HashMap::new();
         let mut nb_records: HashMap<&[u8], i32> = HashMap::new();
-        let file1 = OpenOptions::new()
-            .create(true)
-            .append(true)
-            .open("tests/id1.fa")
-            .expect("cannot open file");
-        let file2 = OpenOptions::new()
-            .create(true)
-            .append(true)
-            .open("tests/id2.fa")
-            .expect("cannot open file");
-        bc_data.insert(b"ACCGTA", vec![file1]);
-        bc_data.insert(b"ATTGTT", vec![file2]);
+
+        let forward = tempfile::tempfile().expect("Cannot create temp file");
+        let unknown = tempfile::tempfile().expect("Cannot create temp file");
+
+        bc_data.insert(b"ACCGTA", vec![forward]);
+        bc_data.insert(b"XXX", vec![unknown]);
 
         assert!(se_fa_demux(
-            &mut records,
-            cmp,
+            "tests/test.fa.gz",
+            niffler::compression::Format::Gzip,
+            niffler::Level::One,
             &bc_data,
             0,
-            out,
             &mut nb_records
         )
         .is_ok());
@@ -653,31 +722,23 @@ mod tests {
 
     #[test]
     fn test_se_fa_demux_m1() {
-        let out = "tests";
-        let (fr, cmp) =
-            read_file("tests/test2.fa.gz").expect("Cannot open file");
-        let mut records = fasta::Reader::new(fr).records();
         let mut bc_data: Barcode = HashMap::new();
         let mut nb_records: HashMap<&[u8], i32> = HashMap::new();
-        let file1 = OpenOptions::new()
-            .create(true)
-            .append(true)
-            .open("tests/id1.fa")
-            .expect("cannot open file");
-        let file2 = OpenOptions::new()
-            .create(true)
-            .append(true)
-            .open("tests/id2.fa")
-            .expect("cannot open file");
-        bc_data.insert(b"ACCGTA", vec![file1]);
-        bc_data.insert(b"ATTGTT", vec![file2]);
+
+        let forward = tempfile::tempfile().expect("Cannot create temp file");
+        let reverse = tempfile::tempfile().expect("Cannot create temp file");
+        let unknown = tempfile::tempfile().expect("Cannot create temp file");
+
+        bc_data.insert(b"ACCGTA", vec![forward]);
+        bc_data.insert(b"ATTGTT", vec![reverse]);
+        bc_data.insert(b"XXX", vec![unknown]);
 
         assert!(se_fa_demux(
-            &mut records,
-            cmp,
+            "tests/test.fa.gz",
+            niffler::compression::Format::Gzip,
+            niffler::Level::One,
             &bc_data,
             1,
-            out,
             &mut nb_records
         )
         .is_ok());
@@ -685,31 +746,23 @@ mod tests {
 
     #[test]
     fn test_se_fa_demux_m2() {
-        let out = "tests";
-        let (fr, cmp) =
-            read_file("tests/test2.fa.gz").expect("Cannot open file");
-        let mut records = fasta::Reader::new(fr).records();
         let mut bc_data: Barcode = HashMap::new();
         let mut nb_records: HashMap<&[u8], i32> = HashMap::new();
-        let file1 = OpenOptions::new()
-            .create(true)
-            .append(true)
-            .open("tests/id1.fa")
-            .expect("cannot open file");
-        let file2 = OpenOptions::new()
-            .create(true)
-            .append(true)
-            .open("tests/id2.fa")
-            .expect("cannot open file");
-        bc_data.insert(b"ACCGTA", vec![file1]);
-        bc_data.insert(b"ATTGTT", vec![file2]);
+
+        let forward = tempfile::tempfile().expect("Cannot create temp file");
+        let reverse = tempfile::tempfile().expect("Cannot create temp file");
+        let unknown = tempfile::tempfile().expect("Cannot create temp file");
+
+        bc_data.insert(b"ACCGTA", vec![forward]);
+        bc_data.insert(b"ATTGTT", vec![reverse]);
+        bc_data.insert(b"XXX", vec![unknown]);
 
         assert!(se_fa_demux(
-            &mut records,
-            cmp,
+            "tests/test.fa.gz",
+            niffler::compression::Format::Gzip,
+            niffler::Level::One,
             &bc_data,
             2,
-            out,
             &mut nb_records
         )
         .is_ok());
@@ -718,31 +771,23 @@ mod tests {
     // se_fq_demux tests ----------------------------------------------------
     #[test]
     fn test_se_fq_demux() {
-        let out = "tests";
-        let (fr, cmp) =
-            read_file("tests/test2.fq.gz").expect("Cannot open file");
-        let mut records = fastq::Reader::new(fr).records();
         let mut bc_data: Barcode = HashMap::new();
         let mut nb_records: HashMap<&[u8], i32> = HashMap::new();
-        let file1 = OpenOptions::new()
-            .create(true)
-            .append(true)
-            .open("tests/id1.fa")
-            .expect("cannot open file");
-        let file2 = OpenOptions::new()
-            .create(true)
-            .append(true)
-            .open("tests/id2.fa")
-            .expect("cannot open file");
-        bc_data.insert(b"ACCGTA", vec![file1]);
-        bc_data.insert(b"ATTGTT", vec![file2]);
+
+        let forward = tempfile::tempfile().expect("Cannot create temp file");
+        let reverse = tempfile::tempfile().expect("Cannot create temp file");
+        let unknown = tempfile::tempfile().expect("Cannot create temp file");
+
+        bc_data.insert(b"ACCGTA", vec![forward]);
+        bc_data.insert(b"ATTGTT", vec![reverse]);
+        bc_data.insert(b"XXX", vec![unknown]);
 
         assert!(se_fq_demux(
-            &mut records,
-            cmp,
+            "tests/test.fq.gz",
+            niffler::compression::Format::Gzip,
+            niffler::Level::One,
             &bc_data,
             0,
-            out,
             &mut nb_records
         )
         .is_ok());
@@ -750,31 +795,23 @@ mod tests {
 
     #[test]
     fn test_se_fq_demux_m1() {
-        let out = "tests";
-        let (fr, cmp) =
-            read_file("tests/test2.fq.gz").expect("Cannot open file");
-        let mut records = fastq::Reader::new(fr).records();
         let mut bc_data: Barcode = HashMap::new();
         let mut nb_records: HashMap<&[u8], i32> = HashMap::new();
-        let file1 = OpenOptions::new()
-            .create(true)
-            .append(true)
-            .open("tests/id1.fa")
-            .expect("cannot open file");
-        let file2 = OpenOptions::new()
-            .create(true)
-            .append(true)
-            .open("tests/id2.fa")
-            .expect("cannot open file");
-        bc_data.insert(b"ACCGTA", vec![file1]);
-        bc_data.insert(b"ATTGTT", vec![file2]);
+
+        let forward = tempfile::tempfile().expect("Cannot create temp file");
+        let reverse = tempfile::tempfile().expect("Cannot create temp file");
+        let unknown = tempfile::tempfile().expect("Cannot create temp file");
+
+        bc_data.insert(b"ACCGTA", vec![forward]);
+        bc_data.insert(b"ATTGTT", vec![reverse]);
+        bc_data.insert(b"XXX", vec![unknown]);
 
         assert!(se_fq_demux(
-            &mut records,
-            cmp,
+            "tests/test.fq.gz",
+            niffler::compression::Format::Gzip,
+            niffler::Level::One,
             &bc_data,
             1,
-            out,
             &mut nb_records
         )
         .is_ok());
@@ -782,31 +819,23 @@ mod tests {
 
     #[test]
     fn test_se_fq_demux_m2() {
-        let out = "tests";
-        let (fr, cmp) =
-            read_file("tests/test2.fq.gz").expect("Cannot open file");
-        let mut records = fastq::Reader::new(fr).records();
         let mut bc_data: Barcode = HashMap::new();
         let mut nb_records: HashMap<&[u8], i32> = HashMap::new();
-        let file1 = OpenOptions::new()
-            .create(true)
-            .append(true)
-            .open("tests/id1.fa")
-            .expect("cannot open file");
-        let file2 = OpenOptions::new()
-            .create(true)
-            .append(true)
-            .open("tests/id2.fa")
-            .expect("cannot open file");
-        bc_data.insert(b"ACCGTA", vec![file1]);
-        bc_data.insert(b"ATTGTT", vec![file2]);
+
+        let forward = tempfile::tempfile().expect("Cannot create temp file");
+        let reverse = tempfile::tempfile().expect("Cannot create temp file");
+        let unknown = tempfile::tempfile().expect("Cannot create temp file");
+
+        bc_data.insert(b"ACCGTA", vec![forward]);
+        bc_data.insert(b"ATTGTT", vec![reverse]);
+        bc_data.insert(b"XXX", vec![unknown]);
 
         assert!(se_fq_demux(
-            &mut records,
-            cmp,
+            "tests/test.fq.gz",
+            niffler::compression::Format::Gzip,
+            niffler::Level::One,
             &bc_data,
             2,
-            out,
             &mut nb_records
         )
         .is_ok());
@@ -818,14 +847,16 @@ mod tests {
         let record =
             fasta::Record::with_attrs("id_str", Some("desc"), b"ATCGCCG");
         let cmp = niffler::compression::Format::Gzip;
-        let file = OpenOptions::new()
-            .create(true)
-            .append(true)
-            .open("tests/mytmp.fa")
-            .expect("cannot open file");
-        assert!((write_to_fa(&file, cmp, &record)).is_ok());
+        let file = tempfile::tempfile().expect("Cannot create temp file");
 
-        let mut fa_records = fasta::Reader::from_file("tests/mytmp.fa")
+        assert!((write_to_fa(&file, cmp, &record, niffler::Level::One)).is_ok());
+
+        let mut tmpfile =
+            tempfile::NamedTempFile::new().expect("Cannot create temp file");
+        writeln!(tmpfile, ">id_str desc\nATCGCCG")
+            .expect("Cannot write to tmp file");
+
+        let mut fa_records = fasta::Reader::from_file(tmpfile)
             .expect("Cannot read file.")
             .records();
 
@@ -846,14 +877,16 @@ mod tests {
             b"QQQQQQQ",
         );
         let cmp = niffler::compression::Format::Gzip;
-        let file = OpenOptions::new()
-            .create(true)
-            .append(true)
-            .open("tests/mytmp.fa")
-            .expect("cannot open file");
-        assert!((write_to_fq(&file, cmp, &record)).is_ok());
+        let file = tempfile::tempfile().expect("Cannot create temp file");
 
-        let mut fa_records = fastq::Reader::from_file("tests/mytmp.fq")
+        assert!((write_to_fq(&file, cmp, &record, niffler::Level::One)).is_ok());
+
+        let mut tmpfile =
+            tempfile::NamedTempFile::new().expect("Cannot create temp file");
+        writeln!(tmpfile, ">id_str desc\nATCGCCG\n+\nQQQQQQQ")
+            .expect("Cannot write to tmp file");
+
+        let mut fa_records = fastq::Reader::from_file(tmpfile)
             .expect("Cannot read file.")
             .records();
 
@@ -868,17 +901,22 @@ mod tests {
     // get_file_type tests -------------------------------------------------
     #[test]
     fn test_get_file_type() {
-        let f1 = "myfile.fa";
-        let f2 = "myfile.fq";
-        let f3 = "myfile.fa.gz";
-        let f4 = "myfile.txt";
-        let f5 = "myfile.txt.xz";
+        let f1 = "tests/reads_1.fa";
+        let f3 = "tests/reads_1.fa.gz";
+        let f5 = "tests/reads_1.fa.xz";
 
-        assert_eq!(get_file_type(f1), Some(FileType::Fasta));
-        assert_eq!(get_file_type(f2), Some(FileType::Fastq));
-        assert_eq!(get_file_type(f3), Some(FileType::Fasta));
-        assert_ne!(get_file_type(f4), Some(FileType::Fastq));
-        assert_eq!(get_file_type(f5), None);
+        assert_eq!(
+            get_file_type(f1).unwrap(),
+            (FileType::Fasta, niffler::compression::Format::No)
+        );
+        assert_eq!(
+            get_file_type(f3).unwrap(),
+            (FileType::Fasta, niffler::compression::Format::Gzip)
+        );
+        assert_eq!(
+            get_file_type(f5).unwrap(),
+            (FileType::Fasta, niffler::compression::Format::Lzma)
+        );
     }
 
     // split_by_tab tests ---------------------------------------------------
@@ -898,4 +936,54 @@ mod tests {
         let mystring = "HelloWorldEarth\nBrianwasthere";
         split_by_tab(mystring).unwrap();
     }
+
+    #[test]
+    fn test_to_niffler_level() {
+        assert_eq!(to_niffler_level(1), niffler::Level::One);
+        assert_eq!(to_niffler_level(2), niffler::Level::Two);
+        assert_eq!(to_niffler_level(3), niffler::Level::Three);
+        assert_eq!(to_niffler_level(4), niffler::Level::Four);
+        assert_eq!(to_niffler_level(5), niffler::Level::Five);
+        assert_eq!(to_niffler_level(6), niffler::Level::Six);
+        assert_eq!(to_niffler_level(7), niffler::Level::Seven);
+        assert_eq!(to_niffler_level(8), niffler::Level::Eight);
+        assert_eq!(to_niffler_level(9), niffler::Level::Nine);
+    }
+
+    #[test]
+    fn test_to_niffler_format() {
+        assert_eq!(
+            to_niffler_format("gz").unwrap(),
+            niffler::compression::Format::Gzip
+        );
+        assert_eq!(
+            to_niffler_format("xz").unwrap(),
+            niffler::compression::Format::Lzma
+        );
+        assert_eq!(
+            to_niffler_format("bz2").unwrap(),
+            niffler::compression::Format::Bzip
+        );
+        assert_eq!(
+            to_niffler_format("txt").unwrap(),
+            niffler::compression::Format::No
+        );
+    }
+
+    #[test]
+    fn test_to_compression_ext() {
+        assert_eq!(
+            to_compression_ext(niffler::compression::Format::Gzip),
+            *".gz"
+        );
+        assert_eq!(
+            to_compression_ext(niffler::compression::Format::Lzma),
+            *".xz"
+        );
+        assert_eq!(
+            to_compression_ext(niffler::compression::Format::Bzip),
+            *".bz2"
+        );
+        assert_eq!(to_compression_ext(niffler::compression::Format::No), *"");
+    }
 }
